AI_AIR_CAP_ZONE = {
    ClassName = "AI_AIR_CAP_ZONE"
}

function AI_AIR_CAP_ZONE:New( AIGroup, PatrolZone, PatrolFloorAltitude, PatrolCeilingAltitude, PatrolMinSpeed, PatrolMaxSpeed, PatrolAltType )
    local self = BASE:Inherit( self, AI_AIR_PATROL_ZONE:New( AIGroup, PatrolZone, PatrolFloorAltitude, PatrolCeilingAltitude, PatrolMinSpeed, PatrolMaxSpeed, PatrolAltType ) ) -- #AI_AIR_CAP_ZONE

    self.Accomplished = false
    self.Engaging = false
  
    self:AddTransition( { "Patrolling", "Engaging" }, "Engage", "Engaging" ) -- FSM_CONTROLLABLE Transition for type #AI_AIR_CAP_ZONE.
    self:AddTransition( "Engaging", "Fired", "Engaging" ) -- FSM_CONTROLLABLE Transition for type #AI_AIR_CAP_ZONE.
    self:AddTransition( "*", "Destroy", "*" ) -- FSM_CONTROLLABLE Transition for type #AI_AIR_CAP_ZONE.
    self:AddTransition( "Engaging", "Abort", "Patrolling" ) -- FSM_CONTROLLABLE Transition for type #AI_AIR_CAP_ZONE.
    self:AddTransition( "Engaging", "Accomplish", "Patrolling" ) -- FSM_CONTROLLABLE Transition for type #AI_AIR_CAP_ZONE.


    return self
    
end
  
  
  --- Set the Engage Zone which defines where the AI will engage bogies. 
  -- @param #AI_AIR_CAP_ZONE self
  -- @param Core.Zone#ZONE EngageZone The zone where the AI is performing CAP.
  -- @return #AI_AIR_CAP_ZONE self
  function AI_AIR_CAP_ZONE:SetEngageZone( EngageZone )
    self:F2()
  
    if EngageZone then  
      self.EngageZone = EngageZone
    else
      self.EngageZone = nil
    end
  end
  
  --- Set the Engage Range when the AI will engage with airborne enemies. 
  -- @param #AI_AIR_CAP_ZONE self
  -- @param #number EngageRange The Engage Range.
  -- @return #AI_AIR_CAP_ZONE self
  function AI_AIR_CAP_ZONE:SetEngageRange( EngageRange )
    self:F2()
  
    if EngageRange then  
      self.EngageRange = EngageRange
    else
      self.EngageRange = nil
    end
  end
  
  --- onafter State Transition for Event Start.
  -- @param #AI_AIR_CAP_ZONE self
  -- @param Wrapper.Controllable#CONTROLLABLE Controllable The Controllable Object managed by the FSM.
  -- @param #string From The From State string.
  -- @param #string Event The Event string.
  -- @param #string To The To State string.
  function AI_AIR_CAP_ZONE:onafterStart( Controllable, From, Event, To )
  
    -- Call the parent Start event handler
    self:GetParent(self).onafterStart( self, Controllable, From, Event, To )
    self:HandleEvent( EVENTS.Dead )
  
  end
  
  
  --- @param AI.AI_CAP#AI_AIR_CAP_ZONE 
  -- @param Wrapper.Group#GROUP EngageGroup
  function AI_AIR_CAP_ZONE.EngageRoute( EngageGroup, Fsm )
  
    EngageGroup:F( { "AI_AIR_CAP_ZONE.EngageRoute:", EngageGroup:GetName() } )
  
    if EngageGroup:IsAlive() then
      Fsm:__Engage( 1 )
    end
  end
  
  
  
  --- @param #AI_AIR_CAP_ZONE self
  -- @param Wrapper.Controllable#CONTROLLABLE Controllable The Controllable Object managed by the FSM.
  -- @param #string From The From State string.
  -- @param #string Event The Event string.
  -- @param #string To The To State string.
  function AI_AIR_CAP_ZONE:onbeforeEngage( Controllable, From, Event, To )
    
    if self.Accomplished == true then
      return false
    end
  end
  
  --- @param #AI_AIR_CAP_ZONE self
  -- @param Wrapper.Controllable#CONTROLLABLE Controllable The Controllable Object managed by the FSM.
  -- @param #string From The From State string.
  -- @param #string Event The Event string.
  -- @param #string To The To State string.
  function AI_AIR_CAP_ZONE:onafterDetected( Controllable, From, Event, To )
    BASE:E("Inside AI_AIR_CAP_ZONE:onafterDetected")
  
    if From ~= "Engaging" then
    
      local Engage = false
    
      for DetectedUnit, Detected in pairs( self.DetectedUnits ) do
      
        local DetectedUnit = DetectedUnit -- Wrapper.Unit#UNIT
        self:T( DetectedUnit )
        if DetectedUnit:IsAlive() and DetectedUnit:IsAir() then
          Engage = true
          break
        end
      end
    
      if Engage == true then
        self:F( 'Detected -> Engaging' )
        self:__Engage( 1 )
      end
    end
  end
  
  
  --- @param #AI_AIR_CAP_ZONE self
  -- @param Wrapper.Controllable#CONTROLLABLE Controllable The Controllable Object managed by the FSM.
  -- @param #string From The From State string.
  -- @param #string Event The Event string.
  -- @param #string To The To State string.
  function AI_AIR_CAP_ZONE:onafterAbort( Controllable, From, Event, To )
    Controllable:ClearTasks()
    self:__Route( 1 )
  end
  
  
  
  
  --- @param #AI_AIR_CAP_ZONE self
  -- @param Wrapper.Controllable#CONTROLLABLE Controllable The Controllable Object managed by the FSM.
  -- @param #string From The From State string.
  -- @param #string Event The Event string.
  -- @param #string To The To State string.
  function AI_AIR_CAP_ZONE:onafterEngage( Controllable, From, Event, To )
  
    if Controllable:IsAlive() then
  
      local EngageRoute = {}
  
      --- Calculate the current route point.
      local CurrentVec2 = self.Controllable:GetVec2()
      
      --TODO: Create GetAltitude function for GROUP, and delete GetUnit(1).
      local CurrentAltitude = self.Controllable:GetUnit(1):GetAltitude()
      local CurrentPointVec3 = POINT_VEC3:New( CurrentVec2.x, CurrentAltitude, CurrentVec2.y )
      local ToEngageZoneSpeed = self.PatrolMaxSpeed
      local CurrentRoutePoint = CurrentPointVec3:WaypointAir( 
          self.PatrolAltType, 
          POINT_VEC3.RoutePointType.TurningPoint, 
          POINT_VEC3.RoutePointAction.TurningPoint, 
          ToEngageZoneSpeed, 
          true 
        )
      
      EngageRoute[#EngageRoute+1] = CurrentRoutePoint
  
      
       --- Find a random 2D point in PatrolZone.
      local ToTargetVec2 = self.PatrolZone:GetRandomVec2()
      self:T2( ToTargetVec2 )
  
      --- Define Speed and Altitude.
      local ToTargetAltitude = math.random( self.EngageFloorAltitude, self.EngageCeilingAltitude )
      local ToTargetSpeed = math.random( self.PatrolMinSpeed, self.PatrolMaxSpeed )
      self:T2( { self.PatrolMinSpeed, self.PatrolMaxSpeed, ToTargetSpeed } )
      
      --- Obtain a 3D @{Point} from the 2D point + altitude.
      local ToTargetPointVec3 = POINT_VEC3:New( ToTargetVec2.x, ToTargetAltitude, ToTargetVec2.y )
      
      --- Create a route point of type air.
      local ToPatrolRoutePoint = ToTargetPointVec3:WaypointAir( 
        self.PatrolAltType, 
        POINT_VEC3.RoutePointType.TurningPoint, 
        POINT_VEC3.RoutePointAction.TurningPoint, 
        ToTargetSpeed, 
        true 
      )
  
      EngageRoute[#EngageRoute+1] = ToPatrolRoutePoint
  
      Controllable:OptionROEOpenFire()
      Controllable:OptionROTEvadeFire()
  
      local AttackTasks = {}
  
      for DetectedUnit, Detected in pairs( self.DetectedUnits ) do
        local DetectedUnit = DetectedUnit -- Wrapper.Unit#UNIT
        self:T( { DetectedUnit, DetectedUnit:IsAlive(), DetectedUnit:IsAir() } )
        if DetectedUnit:IsAlive() and DetectedUnit:IsAir() then
          if self.EngageZone then
            if DetectedUnit:IsInZone( self.EngageZone ) then
              self:F( {"Within Zone and Engaging ", DetectedUnit } )
              AttackTasks[#AttackTasks+1] = Controllable:TaskAttackUnit( DetectedUnit )
            end
          else        
            if self.EngageRange then
              if DetectedUnit:GetPointVec3():Get2DDistance(Controllable:GetPointVec3() ) <= self.EngageRange then
                self:F( {"Within Range and Engaging", DetectedUnit } )
                AttackTasks[#AttackTasks+1] = Controllable:TaskAttackUnit( DetectedUnit )
              end
            else
              AttackTasks[#AttackTasks+1] = Controllable:TaskAttackUnit( DetectedUnit )
            end
          end
        else
          self.DetectedUnits[DetectedUnit] = nil
        end
      end
  
      if #AttackTasks == 0 then
        self:F("No targets found -> Going back to Patrolling")
        self:__Abort( 1 )
        self:__Route( 1 )
        self:SetDetectionActivated()
      else
  
        AttackTasks[#AttackTasks+1] = Controllable:TaskFunction( "AI_AIR_CAP_ZONE.EngageRoute", self )
        EngageRoute[1].task = Controllable:TaskCombo( AttackTasks )
        
        self:SetDetectionDeactivated()
      end
      
      Controllable:Route( EngageRoute, 0.5 )
    
    end
  end
  
  --- @param #AI_AIR_CAP_ZONE self
  -- @param Wrapper.Controllable#CONTROLLABLE Controllable The Controllable Object managed by the FSM.
  -- @param #string From The From State string.
  -- @param #string Event The Event string.
  -- @param #string To The To State string.
  function AI_AIR_CAP_ZONE:onafterAccomplish( Controllable, From, Event, To )
    self.Accomplished = true
    self:SetDetectionOff()
  end
  
  --- @param #AI_AIR_CAP_ZONE self
  -- @param Wrapper.Controllable#CONTROLLABLE Controllable The Controllable Object managed by the FSM.
  -- @param #string From The From State string.
  -- @param #string Event The Event string.
  -- @param #string To The To State string.
  -- @param Core.Event#EVENTDATA EventData
  function AI_AIR_CAP_ZONE:onafterDestroy( Controllable, From, Event, To, EventData )
  
    if EventData.IniUnit then
      self.DetectedUnits[EventData.IniUnit] = nil
    end
  end
  
  --- @param #AI_AIR_CAP_ZONE self
  -- @param Core.Event#EVENTDATA EventData
  function AI_AIR_CAP_ZONE:OnEventDead( EventData )
    self:F( { "EventDead", EventData } )
  
    if EventData.IniDCSUnit then
      if self.DetectedUnits and self.DetectedUnits[EventData.IniUnit] then
        self:__Destroy( 1, EventData )
      end
    end
  end