AI_AIR_PATROL_ZONE = {
    ClassName = "AI_AIR_PATROL_ZONE"
}

function AI_AIR_PATROL_ZONE:New(AIGroup, PatrolZone, PatrolFloorAltitude, PatrolCeilingAltitude, PatrolMinSpeed, PatrolMaxSpeed, PatrolAltType)
    local self = BASE:Inherit(self, AI_AIR:New(AIGroup))

    self.PatrolZone = PatrolZone
    self.PatrolFloorAltitude = PatrolFloorAltitude
    self.PatrolCeilingAltitude = PatrolCeilingAltitude
    self.PatrolMinSpeed = PatrolMinSpeed
    self.PatrolMaxSpeed = PatrolMaxSpeed

    self.PatrolAltType = PatrolAltType or "RADIO"

    self:SetStartState("None")
    self:AddTransition("*", "Stop", "Stopped")
    self:AddTransition( "None", "Start", "Patrolling" )
    self:AddTransition( "Patrolling", "Route", "Patrolling" )
    self:AddTransition( "*", "Detect", "*" )
    self:AddTransition( "*", "Detected", "*" )
    self:AddTransition( "*", "Reset", "Patrolling" )
    self:AddTransition( "*", "Eject", "*" )
    self:AddTransition( "*", "Crash", "Crashed" )
    self:AddTransition( "*", "PilotDead", "*" )

    --self:AddTransition("*", "RTB", "Returning")

    return self
end

function AI_AIR_PATROL_ZONE:SetSpeed(PatrolMinSpeed, PatrolMaxSpeed)
    self:F2( { PatrolMinSpeed, PatrolMaxSpeed } )

    self.PatrolMinSpeed = PatrolMinSpeed
    self.PatrolMaxSpeed = PatrolMaxSpeed
end

function AI_AIR_PATROL_ZONE:SetAltitude( PatrolFloorAltitude, PatrolCeilingAltitude )
    self:F2( { PatrolFloorAltitude, PatrolCeilingAltitude } )
    
    self.PatrolFloorAltitude = PatrolFloorAltitude
    self.PatrolCeilingAltitude = PatrolCeilingAltitude
end

function AI_AIR_PATROL_ZONE:SetDetectionOn()
    self:F2()
  
    self.DetectOn = true
end

function AI_AIR_PATROL_ZONE:SetDetectionOff()
    self:F2()
  
    self.DetectOn = false
end

function AI_AIR_PATROL_ZONE:SetDetectionActivated()
    self:F2()
    
    self:ClearDetectedUnits()
    self.DetectActivated = true
    self:__Detect( -self.DetectInterval )
end

function AI_AIR_PATROL_ZONE:SetDetectionDeactivated()
    self:F2()
    
    self:ClearDetectedUnits()
    self.DetectActivated = false
end


function AI_AIR_PATROL_ZONE:SetRefreshTimeInterval( Seconds )
    self:F2()
  
    if Seconds then  
      self.DetectInterval = Seconds
    else
      self.DetectInterval = 30
    end
end

function AI_AIR_PATROL_ZONE:SetDetectionZone( DetectionZone )
    self:F2()
  
    if DetectionZone then  
      self.DetectZone = DetectionZone
    else
      self.DetectZone = nil
    end
end

function AI_AIR_PATROL_ZONE:GetDetectedUnits()
    self:F2()
  
    return self.DetectedUnits 
end
  
function AI_AIR_PATROL_ZONE:ClearDetectedUnits()
    self:F2()
    self.DetectedUnits = {}
end

function AI_AIR_PATROL_ZONE:onafterStart( Controllable, From, Event, To )
    BASE:E("inside AI_AIR_PATROL_ZONE:onafterStart")
    self:F2()

    self:__Route( 1 ) -- Route to the patrol point. The asynchronous trigger is important, because a spawned group and units takes at least one second to come live.
    self:__Status( 10 ) -- Check status status every 30 seconds.
    self:SetDetectionActivated()
    
    self:HandleEvent( EVENTS.PilotDead, self.OnPilotDead )
    self:HandleEvent( EVENTS.Crash, self.OnCrash )
    self:HandleEvent( EVENTS.Ejection, self.OnEjection )
    
    Controllable:OptionROEHoldFire()
    Controllable:OptionROTVertical()
  
    self.Controllable:OnReSpawn(
      function( PatrolGroup )
        self:E( "ReSpawn" )
        self:__Reset( 1 )
        self:__Route( 5 )
      end
    )
  
    self:SetDetectionOn()
    
end

function AI_AIR_PATROL_ZONE:onbeforeDetect( Controllable, From, Event, To )

    return self.DetectOn and self.DetectActivated
  end
  
  --- @param #AI_AIR_PATROL_ZONE self
  --- @param Wrapper.Controllable#CONTROLLABLE Controllable
  function AI_AIR_PATROL_ZONE:onafterDetect( Controllable, From, Event, To )
    BASE:E("Inside AI_AIR_PATROL_ZONE:onafterDetect")
    local Detected = false
  
    local DetectedTargets = Controllable:GetDetectedTargets()
    for TargetID, Target in pairs( DetectedTargets or {} ) do
        local TargetObject = Target.object
  
        if TargetObject and TargetObject:isExist() and TargetObject.id_ < 50000000 then
  
            local TargetUnit = UNIT:Find( TargetObject )
            local TargetUnitName = TargetUnit:GetName()
        
            if self.DetectionZone then
                if TargetUnit:IsInZone( self.DetectionZone ) then
                    self:T( {"Detected ", TargetUnit } )
                    if self.DetectedUnits[TargetUnit] == nil then
                        self.DetectedUnits[TargetUnit] = true
                    end
                    Detected = true 
                end
            else       
                if self.DetectedUnits[TargetUnit] == nil then
                    self.DetectedUnits[TargetUnit] = true
                end
                Detected = true
            end
        end
    end
  
    self:__Detect( -self.DetectInterval )
    
    if Detected == true then
        self:__Detected( 1.5 )
    end
    
end

function AI_AIR_PATROL_ZONE:_NewPatrolRoute( AIControllable )

    local PatrolZone = AIControllable:GetState( AIControllable, "PatrolZone" ) -- PatrolCore.Zone#AI_AIR_PATROL_ZONE
    PatrolZone:__Route( 1 )
end

function AI_AIR_PATROL_ZONE:onafterRoute( Controllable, From, Event, To )
    self:F2()

    if From == "RTB" then
        return
    end

    if self.Controllable:IsAlive() then
        local PatrolRoute = {}

        if self.Controllable:InAir() == false then
            self:E( "Not in the air, finding route path within PatrolZone" )
            local CurrentVec2 = self.Controllable:GetVec2()
            --TODO: Create GetAltitude function for GROUP, and delete GetUnit(1).
            local CurrentAltitude = self.Controllable:GetUnit(1):GetAltitude()
            local CurrentPointVec3 = POINT_VEC3:New( CurrentVec2.x, CurrentAltitude, CurrentVec2.y )
            local ToPatrolZoneSpeed = self.PatrolMaxSpeed
            local CurrentRoutePoint = CurrentPointVec3:WaypointAir( 
                self.PatrolAltType, 
                POINT_VEC3.RoutePointType.TakeOffParking, 
                POINT_VEC3.RoutePointAction.FromParkingArea, 
                ToPatrolZoneSpeed, 
                true 
              )
            PatrolRoute[#PatrolRoute+1] = CurrentRoutePoint
        else
            self:E( "In the air, finding route path within PatrolZone" )
            local CurrentVec2 = self.Controllable:GetVec2()
            --TODO: Create GetAltitude function for GROUP, and delete GetUnit(1).
            local CurrentAltitude = self.Controllable:GetUnit(1):GetAltitude()
            local CurrentPointVec3 = POINT_VEC3:New( CurrentVec2.x, CurrentAltitude, CurrentVec2.y )
            local ToPatrolZoneSpeed = self.PatrolMaxSpeed
            local CurrentRoutePoint = CurrentPointVec3:WaypointAir( 
                self.PatrolAltType, 
                POINT_VEC3.RoutePointType.TurningPoint, 
                POINT_VEC3.RoutePointAction.TurningPoint, 
                ToPatrolZoneSpeed, 
                true 
              )
            PatrolRoute[#PatrolRoute+1] = CurrentRoutePoint
        end

        local ToTargetVec2 = self.PatrolZone:GetRandomVec2()
        self:T2( ToTargetVec2 )

    --- Define Speed and Altitude.
        local ToTargetAltitude = math.random( self.PatrolFloorAltitude, self.PatrolCeilingAltitude )
        local ToTargetSpeed = math.random( self.PatrolMinSpeed, self.PatrolMaxSpeed )
        self:T2( { self.PatrolMinSpeed, self.PatrolMaxSpeed, ToTargetSpeed } )
    
    --- Obtain a 3D @{Point} from the 2D point + altitude.
        local ToTargetPointVec3 = POINT_VEC3:New( ToTargetVec2.x, ToTargetAltitude, ToTargetVec2.y )
    
    --- Create a route point of type air.
        local ToTargetRoutePoint = ToTargetPointVec3:WaypointAir( 
            self.PatrolAltType, 
            POINT_VEC3.RoutePointType.TurningPoint, 
            POINT_VEC3.RoutePointAction.TurningPoint, 
            ToTargetSpeed, 
            true 
        )

        PatrolRoute[#PatrolRoute+1] = ToTargetRoutePoint
    
    --- Now we're going to do something special, we're going to call a function from a waypoint action at the AIControllable...
        self.Controllable:WayPointInitialize( PatrolRoute )
    
    --- Do a trick, link the NewPatrolRoute function of the PATROLGROUP object to the AIControllable in a temporary variable ...
        self.Controllable:SetState( self.Controllable, "PatrolZone", self )
        self.Controllable:WayPointFunction( #PatrolRoute, 1, "AI_AIR_PATROL_ZONE:_NewPatrolRoute" )

    --- NOW ROUTE THE GROUP!
        self.Controllable:WayPointExecute( 1, 2 )
    end
end

function AI_AIR_PATROL_ZONE:onafterDead()
    self:SetDetectionOff()
    self:SetStatusOff()
end

function AI_AIR_PATROL_ZONE:onafterStatus()
    BASE:E("Inside AI_AIR_PATROL_ZONE:onafterStatus")
    if self.Controllable and self.Controllable:IsAlive() then
  
        local RTB = false
        
        local DistanceFromHomeBase = self.HomeAirbase:GetCoordinate():Get2DDistance( self.Controllable:GetCoordinate() )
        
        if not self:Is( "Holding" ) and not self:Is( "Returning" ) then
            local DistanceFromHomeBase = self.HomeAirbase:GetCoordinate():Get2DDistance( self.Controllable:GetCoordinate() )
          
            if DistanceFromHomeBase > self.DisengageRadius then
                self:E( self.Controllable:GetName() .. " is too far from home base, RTB!" )
                self:Hold( 300 )
                RTB = false
            end
        end    
    
        if not self:Is( "Fuel" ) and not self:Is( "Home" ) then
          
            local Fuel = self.Controllable:GetFuelMin()
          
            -- If the fuel in the controllable is below the treshold percentage,
            -- then send for refuel in case of a tanker, otherwise RTB.
            if Fuel < self.FuelThresholdPercentage then
                BASE:E("Checking fuel")
                if self.TankerName then
                    self:E( self.Controllable:GetName() .. " is out of fuel: " .. Fuel .. " ... Refuelling at Tanker!" )
                    --self:Refuel()
                else
                    self:E( self.Controllable:GetName() .. " is out of fuel: " .. Fuel .. " ... RTB!" )
                    local OldAIControllable = self.Controllable
              
                    local OrbitTask = OldAIControllable:TaskOrbitCircle( math.random( self.PatrolFloorAltitude, self.PatrolCeilingAltitude ), self.PatrolMinSpeed )
                    local TimedOrbitTask = OldAIControllable:TaskControlled( OrbitTask, OldAIControllable:TaskCondition(nil,nil,nil,nil,self.OutOfFuelOrbitTime,nil ) )
                    OldAIControllable:SetTask( TimedOrbitTask, 10 )
        
                    --self:Fuel()
                    RTB = true
                end
            else
            end
        end
        
        -- TODO: Check GROUP damage function.
        local Damage = self.Controllable:GetLife()
        local InitialLife = self.Controllable:GetLife0()
        
        -- If the group is damaged, then RTB.
        -- Note that a group can consist of more units, so if one unit is damaged of a group, the mission may continue.
        -- The damaged unit will RTB due to DCS logic, and the others will continue to engage.
        if ( Damage / InitialLife ) < self.PatrolDamageThreshold then
            self:E( self.Controllable:GetName() .. " is damaged: " .. Damage .. " ... RTB!" )
            self:Damaged()
            RTB = true
            self:SetStatusOff()
        end
        
        -- Check if planes went RTB and are out of control.
        -- We only check if planes are out of control, when they are in duty.
        if self.Controllable:HasTask() == false then
            if not self:Is( "Started" ) and 
                not self:Is( "Stopped" ) and
                not self:Is( "Fuel" ) and 
                not self:Is( "Damaged" ) and 
                not self:Is( "Home" ) then
                if self.IdleCount >= 10 then
                    if Damage ~= InitialLife then
                        self:Damaged()
                    else  
                        self:E( self.Controllable:GetName() .. " control lost! " )
                
                        self:LostControl()
                    end
                else
                    self.IdleCount = self.IdleCount + 1
                end
            end
        else
            self.IdleCount = 0
        end
    
        if RTB == true then
            self:__RTB( self.TaskDelay )
        end
    
        if not self:Is("Home") then
            self:__Status( 10 )
        end
        
    end
end
  --- @param #AI_AIR_PATROL_ZONE self
  -- @param Core.Event#EVENTDATA EventData
function AI_AIR_PATROL_ZONE:OnCrash( EventData )
  
    if self.Controllable:IsAlive() and EventData.IniDCSGroupName == self.Controllable:GetName() then
        if #self.Controllable:GetUnits() == 1 then
            self:__Crash( 1, EventData )
        end
    end
end
  
  --- @param #AI_AIR_PATROL_ZONE self
  -- @param Core.Event#EVENTDATA EventData
function AI_AIR_PATROL_ZONE:OnEjection( EventData )
  
    if self.Controllable:IsAlive() and EventData.IniDCSGroupName == self.Controllable:GetName() then
      self:__Eject( 1, EventData )
    end
end
  
  --- @param #AI_AIR_PATROL_ZONE self
  -- @param Core.Event#EVENTDATA EventData
function AI_AIR_PATROL_ZONE:OnPilotDead( EventData )
  
    if self.Controllable:IsAlive() and EventData.IniDCSGroupName == self.Controllable:GetName() then
      self:__PilotDead( 1, EventData )
    end
end